import type { AppContext } from 'vue';
import type { Component } from 'vue';
import type { ComponentResolver } from 'unplugin-vue-components';
import { ComputedRef } from 'vue';
import type { ExtractPropTypes } from 'vue';
import { InjectionKey } from 'vue';
import { isNumber } from '@vueuse/core';
import { isString } from '@vue/shared';
import type { Plugin as Plugin_2 } from 'vue';
import type { PropType } from 'vue';
import { Ref } from 'vue';

export declare function addUnit(value: string | number, defaultUnit?: string): string;

/**
 * @description Build prop. It can better optimize prop types
 * @description 生成 prop，能更好地优化类型
 * @example
 // limited options
 // the type will be PropType<'light' | 'dark'>
 buildProp({
 type: String,
 values: ['light', 'dark'],
 } as const)
 * @example
 // limited options and other types
 // the type will be PropType<'small' | 'large' | number>
 buildProp({
 type: [String, Number],
 values: ['small', 'large'],
 validator: (val: unknown): val is number => typeof val === 'number',
 } as const)
 @link see more: https://github.com/element-plus/element-plus/pull/3341
 */
export declare const buildProp: <Type = never, Value = never, Validator = never, Default extends InPropMergeType<Type, Value, Validator> = never, Required_1 extends boolean = false>(prop: InPropInput<Type, Value, Validator, Default, Required_1>, key?: string) => InPropFinalized<Type, Value, Validator, Default, Required_1>;

export declare const buildProps: <Props extends Record<string, NativePropType | {
    __epPropKey: true;
} | InPropInput<any, any, any, any, any>>>(props: Props) => { [K in keyof Props]: IfInProp<Props[K], Props[K], IfNativePropType<Props[K], Props[K], InPropConvert<Props[K]>>>; };

export declare type ComponentStylingHook<P, R = string> = (props: P, ns?: UseNamespaceReturn) => ComputedRef<R>;

/**
 *
 * @param func
 * @param delay
 * @returns
 */
export declare function debounce<T extends (...args: any[]) => void>(func: T, delay: number): (...args: Parameters<T>) => void;

export declare function debugWarn(err: Error): void;

export declare function debugWarn(scope: string, message: string): void;

export declare const defaultNamespace = "in";

export declare const definePropType: <T>(val: any) => PropType<T>;

/**
 * Extract the type of a single prop
 *
 * 提取单个 prop 的参数类型
 *
 * @example
 * ExtractPropType<{ type: StringConstructor }> => string | undefined
 * ExtractPropType<{ type: StringConstructor, required: true }> => string
 * ExtractPropType<{ type: BooleanConstructor }> => boolean
 */
export declare type ExtractPropType<T extends object> = Value<ExtractPropTypes<{
    key: T;
}>>;

export declare const generateId: (length?: number) => string;

/**
 * Determine if it is `InProp`
 */
export declare type IfInProp<T, Y, N> = T extends {
    [inPropKey]: true;
} ? Y : N;

export declare type IfNativePropType<T, Y, N> = [T] extends [NativePropType] ? Y : N;

export declare type IfNever<T, Y = true, N = false> = [T] extends [never] ? Y : N;

export declare type IfUnknown<T, Y, N> = [unknown] extends [T] ? Y : N;

/**
 * output prop `buildProp` or `buildProps`.
 *
 * prop 输出参数。
 *
 * @example
 * InProp<'a', 'b', true>
 * ⬇️
 * {
 readonly type: PropType<"a">;
 readonly required: true;
 readonly validator: ((val: unknown) => boolean) | undefined;
 readonly default: "b";
 __epPropKey: true;
 }
 */
export declare type InProp<Type, Default, Required> = {
    readonly type: PropType<Type>;
    readonly required: [Required] extends [true] ? true : false;
    readonly validator: ((val: unknown) => boolean) | undefined;
    [inPropKey]: true;
} & IfNever<Default, unknown, {
    readonly default: Default;
}>;

/**
 * Converting input to output.
 *
 * 将输入转换为输出
 */
export declare type InPropConvert<Input> = Input extends InPropInput<infer Type, infer Value, infer Validator, any, infer Required> ? InPropFinalized<Type, Value, Validator, Input['default'], Required> : never;

/**
 * Finalized conversion output
 *
 * 最终转换 InProp
 */
export declare type InPropFinalized<Type, Value, Validator, Default, Required> = InProp<InPropMergeType<Type, Value, Validator>, UnknownToNever<Default>, Required>;

/**
 * input prop `buildProp` or `buildProps` (constraints)
 *
 * prop 输入参数（约束）
 *
 * @example
 * InPropInput<StringConstructor, 'a', never, never, true>
 * ⬇️
 * {
 type?: StringConstructor | undefined;
 required?: true | undefined;
 values?: readonly "a"[] | undefined;
 validator?: ((val: any) => boolean) | ((val: any) => val is never) | undefined;
 default?: undefined;
 }
 */
export declare interface InPropInput<Type, Value, Validator, Default extends InPropMergeType<Type, Value, Validator>, Required extends boolean> {
    type?: Type;
    required?: Required;
    values?: readonly Value[];
    validator?: ((val: any) => val is Validator) | ((val: any) => boolean);
    default?: InPropInputDefault<Required, Default>;
}

/**
 * Handling default values for input (constraints)
 *
 * 处理输入参数的默认值（约束）
 */
export declare type InPropInputDefault<Required extends boolean, Default> = Required extends true ? never : Default extends Record<string, unknown> | Array<any> ? () => Default : (() => Default) | Default;

export declare const inPropKey = "__epPropKey";

/**
 * Merge Type, Value, Validator types
 * 合并 Type、Value、Validator 的类型
 *
 * @example
 * InPropMergeType<StringConstructor, '1', 1> =>  1 | "1" // ignores StringConstructor
 * InPropMergeType<StringConstructor, never, number> =>  string | number
 */
export declare type InPropMergeType<Type, Value, Validator> = IfNever<UnknownToNever<Value>, ResolvePropType<Type>, never> | UnknownToNever<Value> | UnknownToNever<Validator>;

export declare const isElement: (e: unknown) => e is Element;

export declare const isEmpty: (val: unknown) => boolean;

export declare const isEpProp: (val: unknown) => val is InProp<any, any, any>;

export { isNumber }

export declare const isPropAbsent: (prop: unknown) => prop is null | undefined;

export declare const isSSR: boolean;

export { isString }

export declare const isStringNumber: (val: string) => boolean;

export declare const isUndefined: (val: any) => val is undefined;

export declare function kebabCase(key: string): string;

export declare const namespaceContextKey: InjectionKey<Ref<string | undefined>>;

/**
 * Native prop types, e.g: `BooleanConstructor`, `StringConstructor`, `null`, `undefined`, etc.
 *
 * 原生 prop `类型，BooleanConstructor`、`StringConstructor`、`null`、`undefined` 等
 */
export declare type NativePropType = ((...args: any) => any) | {
    new (...args: any): any;
} | undefined | null;

/**
 * 移除事件监听
 * @param {*} element
 * @param {*} event
 * @param {*} handler
 * @param {*} useCapture
 */
export declare function off(element: {
    removeEventListener: (arg0: any, arg1: any, arg2: boolean) => void;
}, event: any, handler: any, useCapture?: any): void;

/**
 * 新增事件监听
 * @param {*} element
 * @param {*} event
 * @param {*} handler
 * @param {*} useCapture
 */
export declare function on(element: {
    addEventListener: (arg0: any, arg1: any, arg2: boolean) => void;
}, event: any, handler: any, useCapture?: any): void;

export declare function PlayResolver(options?: PlayResolverOptions): ComponentResolver;

export declare interface PlayResolverOptions {
    importStyle?: boolean | 'less';
    from?: string;
}

declare type Refs<T extends object> = {
    [K in keyof T]: Ref<T[K]>;
};

/**
 * Extracts types via `ExtractPropTypes`, accepting `PropType<T>`, `XXXConstructor`, `never`...
 *
 * 通过 `ExtractPropTypes` 提取类型，接受 `PropType<T>`、`XXXConstructor`、`never`...
 *
 * @example
 * ResolvePropType<BooleanConstructor> => boolean
 * ResolvePropType<PropType<T>> => T
 **/
export declare type ResolvePropType<T> = IfNever<T, never, ExtractPropType<{
    type: WritableArray<T>;
    required: true;
}>>;

export declare type SFCInstallWithContext<T> = SFCWithInstall<T> & {
    _context: AppContext | null;
};

export declare type SFCWithInstall<T> = T & Plugin_2;

export declare const singleton: <T extends {}>(className: T) => T;

/**
 *
 * @param func
 * @param delay
 * @returns
 */
export declare function throttle<T extends (...args: any[]) => void>(func: T, delay: number): (...args: Parameters<T>) => void;

/**
 * Capitalize the first letter of a string
 *
 * @param {string} str - The string to capitalize
 * @return {string} The capitalized string
 */
export declare const toCapitalize: (str: string) => string;

/**
 * Convert camelCase to kebab-case
 *
 * @param {string} str - The string to convert
 * @return {string} The converted string
 */
export declare const toKebabCase: (str: string) => string;

/**
 * Convert kebab-case to lower camelCase
 *
 * @param {string} str - The string to convert
 * @return {string} The converted string
 */
export declare const toLowerCamelCase: (str: string) => string;

/**
 * Convert kebab-case to upper camelCase
 *
 * @param {string} str - The string to convert
 * @return {string} The converted string
 */
export declare const toUpperCamelCase: (str: string) => string;

export declare type UnknownToNever<T> = IfUnknown<T, never, T>;

declare type UnRefs = <T extends object>(object: Refs<T>) => T;

export declare const unRefs: UnRefs;

export declare const useGetDerivedNamespace: (namespaceOverrides?: Ref<string | undefined>) => ComputedRef<string>;

export declare const useNamespace: (block: string, namespaceOverrides?: Ref<string | undefined>) => {
    namespace: ComputedRef<string>;
    b: (blockSuffix?: string) => string;
    e: (element?: string) => string;
    m: (modifier?: string) => string;
    be: (blockSuffix?: string, element?: string) => string;
    em: (element?: string, modifier?: string) => string;
    bm: (blockSuffix?: string, modifier?: string) => string;
    bem: (blockSuffix?: string, element?: string, modifier?: string) => string;
    is: {
        (name: string, state: boolean | undefined): string;
        (name: string): string;
    };
    cssVar: (object: Record<string, string>) => Record<string, string>;
    cssStyle: (object: Record<string, string>) => Record<string, string>;
    cssVarName: (name: string) => string;
    cssVarBlock: (object: Record<string, string | number>) => Record<string, string | number>;
    cssVarBlockName: (name: string) => string;
    getCssVar: (name: string) => string;
    getCssVarBlock: (name: string, blockOverrides?: string) => string;
};

export declare type UseNamespaceReturn = ReturnType<typeof useNamespace>;

declare type Value<T> = T[keyof T];

export declare function withInstall<T extends Component>(comp: T, name: string): SFCWithInstall<T>;

export declare const withInstallFunction: <T>(fn: T, name: string) => SFCInstallWithContext<T>;

export declare type Writable<T> = {
    -readonly [P in keyof T]: T[P];
};

export declare type WritableArray<T> = T extends readonly any[] ? Writable<T> : T;

export { }
